@@ -12,7 +12,7 @@ use warnings;
 # It is provided simply to keep CPAN happy:
 #   cpan -i Tangence
 
-our $VERSION = '0.20_001';
+our $VERSION = '0.01';
 
 =head1 NAME
 
@@ -35,4 +35,5 @@ Paul Evans <leonerd@leonerd.org.uk>
 
 =cut
 
-0x55AA;
+# Keep perl happy; keep Britain tidy
+1;
