@@ -1,48 +1,28 @@
 #  You may distribute under the terms of either the GNU General Public License
 #  or the Artistic License (the same terms as Perl itself)
 #
-#  (C) Paul Evans, 2010-2014 -- leonerd@leonerd.org.uk
+#  (C) Paul Evans, 2010 -- leonerd@leonerd.org.uk
 
 package Tangence::Object;
 
 use strict;
-use warnings;
-
-our $VERSION = '0.20_001';
 
 use Carp;
 
 use Tangence::Constants;
+use Tangence::Metacode;
 
-use Tangence::Types;
+our %METHODS = (
+);
 
-use Tangence::Class;
-
-Tangence::Class->declare(
-   __PACKAGE__,
-
-   events => {
-      destroy => {
-         args => [],
-      },
+our %EVENTS = (
+   destroy => {
+      args => [],
    },
 );
 
-=head1 NAME
-
-C<Tangence::Object> - base class for accessible objects in a C<Tangence> server
-
-=head1 DESCRIPTION
-
-This class acts as a base class for the accessible objects in a L<Tangence>
-server. All the objects actually created and made accessible to clients will
-be subclasses of this one, including internally-created objects such as
-L<Tangence::Registry>.
-
-These objects are not directly constructed by calling the C<new> class method;
-instead the C<Tangence::Registry> should be used to construct one.
-
-=cut
+our %PROPS = (
+);
 
 sub new
 {
@@ -52,17 +32,18 @@ sub new
    defined( my $id = delete $args{id} ) or croak "Need a id";
    my $registry = delete $args{registry} or croak "Need a registry";
 
+   Tangence::Metacode::init_class( $class ) unless do { no strict 'refs'; defined &{"${class}::_has_Tangence"} };
+
    my $self = bless {
       id => $id,
       registry => $registry,
-      meta => $args{meta} || Tangence::Class->for_perlname( $class ),
 
       event_subs => {},   # {$event} => [ @cbs ]
 
       properties => {}, # {$prop} => [ $value, \@callbacks ]
    }, $class;
 
-   my $properties = $self->class->properties;
+   my $properties = $self->can_property();
    foreach my $prop ( keys %$properties ) {
       $self->_new_property( $prop, $properties->{$prop} );
    }
@@ -75,16 +56,16 @@ sub _new_property
    my $self = shift;
    my ( $prop, $pdef ) = @_;
 
-   my $dim = $pdef->dimension;
+   my $dim = $pdef->{dim};
 
    my $initial;
 
    if( my $code = $self->can( "init_prop_$prop" ) ) {
       $initial = $code->( $self );
    }
-
+ 
    elsif( $dim == DIM_SCALAR ) {
-      $initial = $pdef->type->default_value;
+      $initial = undef;
    }
    elsif( $dim == DIM_HASH ) {
       $initial = {};
@@ -102,19 +83,11 @@ sub _new_property
    $self->{properties}->{$prop} = [ $initial, [] ];
 }
 
-=head1 METHODS
-
-=cut
-
-=head2 $obj->destroy
-
-Requests that the object destroy itself, informing all clients that are aware
-of it. Once they all report that they have dropped the object, the object is
-deconstructed for real.
-
-Not to be confused with Perl's own C<DESTROY> method.
-
-=cut
+sub DESTROY
+{
+   my $self = shift;
+   $self->destroy if defined $self->{registry};
+}
 
 sub destroy
 {
@@ -154,38 +127,18 @@ sub _destroy_really
    $self->{destroyed} = 1;
 }
 
-=head2 $id = $obj->id
-
-Returns the object's C<Tangence> ID number
-
-=cut
-
 sub id
 {
    my $self = shift;
    return $self->{id};
 }
 
-=head2 $description = $obj->describe
-
-Returns a textual description of the object, for internal debugging purposes.
-Subclasses are encouraged to override this method to return something more
-descriptive within their domain of interest
-
-=cut
-
 sub describe
 {
    my $self = shift;
    return ref $self;
 }
 
-=head2 $registry = $obj->registry
-
-Returns the L<Tangence::Registry> that constructed this object.
-
-=cut
-
 sub registry
 {
    my $self = shift;
@@ -213,70 +166,127 @@ sub smash
    } @keys };
 }
 
-=head2 $class = $obj->class
-
-Returns the L<Tangence::Meta::Class> object representing the class of this
-object.
-
-=cut
-
-sub class
-{
-   my $self = shift;
-   return ref $self ? $self->{meta} : Tangence::Class->for_perlname( $self );
-}
-
-=head2 $method = $obj->can_method( $name )
-
-Returns the L<Tangence::Meta::Method> object representing the named method, or
-C<undef> if no such method exists.
-
-=cut
-
 sub can_method
 {
    my $self = shift;
-   return $self->class->method( @_ );
+   my ( $method, $class ) = @_;
+
+   $class ||= ( ref $self || $self );
+
+   my %methods = do { no strict 'refs'; %{$class."::METHODS"} };
+
+   return $methods{$method} if defined $method and exists $methods{$method};
+
+   my @isa = do { no strict 'refs'; @{$class."::ISA"} };
+
+   foreach my $superclass ( @isa ) {
+      my $m = $self->can_method( $method, $superclass );
+      if( defined $method ) {
+         return $m if $m;
+      }
+      else {
+         exists $methods{$_} or $methods{$_} = $m->{$_} for keys %$m;
+      }
+   }
+
+   return \%methods unless defined $method;
+   return undef;
 }
 
-=head2 $event = $obj->can_event( $name )
-
-Returns the L<Tangence::Meta::Event> object representing the named event, or
-C<undef> if no such event exists.
-
-=cut
-
 sub can_event
 {
    my $self = shift;
-   return $self->class->event( @_ );
+   my ( $event, $class ) = @_;
+
+   $class ||= ( ref $self || $self );
+
+   my %events = do { no strict 'refs'; %{$class."::EVENTS"} };
+
+   return $events{$event} if defined $event and exists $events{$event};
+
+   my @isa = do { no strict 'refs'; @{$class."::ISA"} };
+
+   foreach my $superclass ( @isa ) {
+      my $e = $self->can_event( $event, $superclass );
+      if( defined $event ) {
+         return $e if $e;
+      }
+      else {
+         exists $events{$_} or $events{$_} = $e->{$_} for keys %$e;
+      }
+   }
+
+   return \%events unless defined $event;
+   return undef;
 }
 
-=head2 $property = $obj->can_property( $name )
-
-Returns the L<Tangence::Meta::Property> object representing the named
-property, or C<undef> if no such property exists.
-
-=cut
-
 sub can_property
 {
    my $self = shift;
-   return $self->class->property( @_ );
+   my ( $prop, $class ) = @_;
+
+   $class ||= ( ref $self || $self );
+
+   my %props = do { no strict 'refs'; %{$class."::PROPS"} };
+
+   return $props{$prop} if defined $prop and exists $props{$prop};
+
+   my @isa = do { no strict 'refs'; @{$class."::ISA"} };
+
+   foreach my $superclass ( @isa ) {
+      my $p = $self->can_property( $prop, $superclass );
+      if( defined $prop ) {
+         return $p if $p;
+      }
+      else {
+         exists $props{$_} or $props{$_} = $p->{$_} for keys %$p;
+      }
+   }
+
+   return \%props unless defined $prop;
+   return undef;
 }
 
 sub smashkeys
 {
    my $self = shift;
-   return $self->class->smashkeys;
+   my ( $class ) = @_;
+
+   $class ||= ( ref $self || $self );
+
+   my %props = do { no strict 'refs'; %{$class."::PROPS"} };
+
+   my %smash;
+
+   $props{$_}->{smash} and $smash{$_} = 1 for keys %props;
+
+   my @isa = do { no strict 'refs'; @{$class."::ISA"} };
+
+   foreach my $superclass ( @isa ) {
+      my $supkeys = $self->smashkeys( $superclass );
+
+      # Merge keys we don't yet have
+      $smash{$_} = 1 for keys %$supkeys;
+   }
+
+   return \%smash;
 }
 
-=head2 $obj->fire_event( $event, @args )
+sub introspect
+{
+   my $self = shift;
 
-Fires the named event on the object. Each event subscription function will be
-invoked with the given arguments.
+   my $class = ( ref $self || $self );
 
-=cut
+   my $ret = {
+      methods    => $self->can_method(),
+      events     => $self->can_event(),
+      properties => $self->can_property(),
+      isa        => [ $class, do { no strict 'refs'; @{$class."::ISA"} } ],
+   };
+
+   return $ret;
+}
 
 sub fire_event
 {
@@ -287,26 +297,11 @@ sub fire_event
 
    $self->can_event( $event ) or croak "$self has no event $event";
 
-   my $sublist = $self->{event_subs}->{$event} or return;
-
-   foreach my $cb ( @$sublist ) {
-      $cb->( $self, @args );
+   foreach my $cb ( @{ $self->{event_subs}->{$event} } ) {
+      $cb->( @args );
    }
 }
 
-=head2 $id = $obj->subscribe_event( $event, $callback )
-
-Subscribes an event-handling callback CODE ref to the named event. When the
-event is fired by C<fire_event> this callback will be invoked, being passed
-the object reference and the event's arguments.
-
- $callback->( $obj, @args )
-
-Returns an opaque ID value that can be used to remove this subscription by
-calling C<unsubscribe_event>.
-
-=cut
-
 sub subscribe_event
 {
    my $self = shift;
@@ -322,19 +317,12 @@ sub subscribe_event
    return $ref + 0; # force numeric context
 }
 
-=head2 $obj->unsubscribe_event( $event, $id )
-
-Removes an event-handling callback previously registered with
-C<subscribe_event>.
-
-=cut
-
 sub unsubscribe_event
 {
    my $self = shift;
    my ( $event, $id ) = @_;
 
-   my $sublist = $self->{event_subs}->{$event} or return;
+   my $sublist = $self->{event_subs}->{$event};
 
    my $index;
    for( $index = 0; $index < @$sublist; $index++ ) {
@@ -344,48 +332,6 @@ sub unsubscribe_event
    splice @$sublist, $index, 1, ();
 }
 
-=head2 $id = $obj->watch_property( $prop, %callbacks )
-
-Watches a named property for changes, registering a set of callback functions
-to be invoked when the property changes in certain ways. The set of callbacks
-required depends on the dimension of the property being watched.
-
-For all property types:
-
- $on_set->( $obj, $value )
-
-For hash properties:
-
- $on_add->( $obj, $key, $value )
- $on_del->( $obj, $key )
-
-For queue properties:
-
- $on_push->( $obj, @values )
- $on_shift->( $obj, $count )
-
-For array properties:
-
- $on_push->( $obj, @values )
- $on_shift->( $obj, $count )
- $on_splice->( $obj, $index, $count, @values )
- $on_move->( $obj, $index, $delta )
-
-For objset properties:
-
- $on_add->( $obj, $added_object )
- $on_del->( $obj, $deleted_object_id )
-
-Alternatively, a single callback may be installed that is invoked after any
-change of the property, being passed the new value entirely:
-
- $on_updated->( $obj, $value )
-
-Returns an opaque ID value that can be used to remove this watch by calling
-C<unwatch_property>.
-
-=cut
-
 sub watch_property
 {
    my $self = shift;
@@ -403,7 +349,7 @@ sub watch_property
       $callbacks->{on_updated} = $on_updated;
    }
    else {
-      foreach my $name ( @{ CHANGETYPES->{$pdef->dimension} } ) {
+      foreach my $name ( @{ CHANGETYPES->{$pdef->{dim}} } ) {
          ref( $callbacks->{$name} = delete $callbacks{$name} ) eq "CODE"
             or croak "Expected '$name' as a CODE ref";
       }
@@ -413,25 +359,18 @@ sub watch_property
 
    push @$watchlist, $callbacks;
 
-   $on_updated->( $self, $self->{properties}->{$prop}->[0] ) if $on_updated;
+   $on_updated->( $self->{properties}->{$prop}->[0] ) if $on_updated;
 
    my $ref = \@{$watchlist}[$#$watchlist];  # reference to last element
    return $ref + 0; # force numeric context
 }
 
-=head2 $obj->unwatch_property( $prop, $id )
-
-Removes the set of callback functions previously registered with
-C<watch_property>.
-
-=cut
-
 sub unwatch_property
 {
    my $self = shift;
    my ( $prop, $id ) = @_;
 
-   my $watchlist = $self->{properties}->{$prop}->[1] or return;
+   my $watchlist = $self->{properties}->{$prop}->[1];
 
    my $index;
    for( $index = 0; $index < @$watchlist; $index++ ) {
@@ -455,12 +394,12 @@ sub handle_request_CALL
    my $m = "method_$method";
    $self->can( $m ) or die "Object cannot run method $method\n";
 
-   my @args = map { $_->unpack_value( $message ) } $mdef->argtypes;
+   my @args = $message->unpack_all_typed( $mdef->{args} );
 
    my $result = $self->$m( $ctx, @args );
 
-   my $response = Tangence::Message->new( $ctx->stream, MSG_RESULT );
-   $mdef->ret->pack_value( $response, $result ) if $mdef->ret;
+   my $response = Tangence::Message->new( $ctx->connection, MSG_RESULT );
+   $response->pack_typed( $mdef->{ret}, $result ) if $mdef->{ret};
 
    return $response;
 }
@@ -472,14 +411,10 @@ sub generate_message_EVENT
 
    my $edef = $self->can_event( $event ) or die "Object cannot respond to event $event";
 
-   my $response = Tangence::Message->new( $conn, MSG_EVENT )
+   return Tangence::Message->new( $conn, MSG_EVENT )
       ->pack_int( $self->id )
-      ->pack_str( $event );
-
-   my @argtypes = $edef->argtypes;
-   $argtypes[$_]->pack_value( $response, $args[$_] ) for 0..$#argtypes;
-
-   return $response;
+      ->pack_str( $event )
+      ->pack_all_typed( $edef->{args}, @args );
 }
 
 sub handle_request_GETPROP
@@ -496,42 +431,8 @@ sub handle_request_GETPROP
 
    my $result = $self->$m();
 
-   my $response = Tangence::Message->new( $ctx->stream, MSG_RESULT );
-   $pdef->overall_type->pack_value( $response, $result );
-
-   return $response;
-}
-
-sub handle_request_GETPROPELEM
-{
-   my $self = shift;
-   my ( $ctx, $message ) = @_;
-
-   my $prop = $message->unpack_str();
-
-   my $pdef = $self->can_property( $prop ) or die "Object does not have property $prop";
-   my $dim = $pdef->dimension;
-
-   my $m = "get_prop_$prop";
-   $self->can( $m ) or die "Object cannot get property $prop\n";
-
-   my $result;
-   if( $dim == DIM_QUEUE or $dim == DIM_ARRAY ) {
-      my $idx = $message->unpack_int();
-      $result = $self->$m()->[$idx];
-   }
-   elsif( $dim == DIM_HASH ) {
-      my $key = $message->unpack_str();
-      $result = $self->$m()->{$key};
-   }
-   else {
-      die "Property $prop cannot fetch elements";
-   }
-
-   my $response = Tangence::Message->new( $ctx->stream, MSG_RESULT );
-   $pdef->type->pack_value( $response, $result );
-
-   return $response;
+   return Tangence::Message->new( $ctx->connection, MSG_RESULT )
+      ->pack_any( $result );
 }
 
 sub handle_request_SETPROP
@@ -540,11 +441,10 @@ sub handle_request_SETPROP
    my ( $ctx, $message ) = @_;
 
    my $prop  = $message->unpack_str();
+   my $value = $message->unpack_any();
 
    my $pdef = $self->can_property( $prop ) or die "Object does not have property $prop\n";
 
-   my $value = $pdef->type->unpack_value( $message );
-
    my $m = "set_prop_$prop";
    $self->can( $m ) or die "Object cannot set property $prop\n";
 
@@ -559,20 +459,16 @@ sub generate_message_UPDATE
    my ( $conn, $prop, $how, @args ) = @_;
 
    my $pdef = $self->can_property( $prop ) or die "Object does not have property $prop\n";
-   my $dim = $pdef->dimension;
+   my $dim = $pdef->{dim};
 
    my $message = Tangence::Message->new( $conn, MSG_UPDATE )
       ->pack_int( $self->id )
-      ->pack_str( $prop );
-   TYPE_U8->pack_value( $message, $how );
+      ->pack_str( $prop )
+      ->pack_typed( "u8", $how );
 
    my $dimname = DIMNAMES->[$dim];
-   if( $how == CHANGE_SET ) {
-      my ( $value ) = @args;
-      $pdef->overall_type->pack_value( $message, $value );
-   }
-   elsif( my $code = $self->can( "_generate_message_UPDATE_$dimname" ) ) {
-      $code->( $self, $message, $how, $pdef, @args );
+   if( my $code = $self->can( "_generate_message_UPDATE_$dimname" ) ) {
+      $code->( $self, $message, $how, $pdef->{type}, @args );
    }
    else {
       croak "Unrecognised property dimension $dim for $prop";
@@ -584,20 +480,30 @@ sub generate_message_UPDATE
 sub _generate_message_UPDATE_scalar
 {
    my $self = shift;
-   my ( $message, $how, $pdef, @args ) = @_;
+   my ( $message, $how, $type, @args ) = @_;
 
-   croak "Change type $how is not valid for a scalar property";
+   if( $how == CHANGE_SET ) {
+      my ( $value ) = @args;
+      $message->pack_typed( $type, $value );
+   }
+   else {
+      croak "Change type $how is not valid for a scalar property";
+   }
 }
 
 sub _generate_message_UPDATE_hash
 {
    my $self = shift;
-   my ( $message, $how, $pdef, @args ) = @_;
+   my ( $message, $how, $type, @args ) = @_;
 
-   if( $how == CHANGE_ADD ) {
+   if( $how == CHANGE_SET ) {
+      my ( $value ) = @args;
+      $message->pack_typed( "dict($type)", $value );
+   }
+   elsif( $how == CHANGE_ADD ) {
       my ( $key, $value ) = @args;
       $message->pack_str( $key );
-      $pdef->type->pack_value( $message, $value );
+      $message->pack_typed( $type, $value );
    }
    elsif( $how == CHANGE_DEL ) {
       my ( $key ) = @args;
@@ -611,10 +517,14 @@ sub _generate_message_UPDATE_hash
 sub _generate_message_UPDATE_queue
 {
    my $self = shift;
-   my ( $message, $how, $pdef, @args ) = @_;
+   my ( $message, $how, $type, @args ) = @_;
 
-   if( $how == CHANGE_PUSH ) {
-      $message->pack_all_sametype( $pdef->type, @args );
+   if( $how == CHANGE_SET ) {
+      my ( $value ) = @args;
+      $message->pack_typed( "list($type)", $value );
+   }
+   elsif( $how == CHANGE_PUSH ) {
+      $message->pack_all_sametype( $type, @args );
    }
    elsif( $how == CHANGE_SHIFT ) {
       my ( $count ) = @args;
@@ -628,10 +538,14 @@ sub _generate_message_UPDATE_queue
 sub _generate_message_UPDATE_array
 {
    my $self = shift;
-   my ( $message, $how, $pdef, @args ) = @_;
+   my ( $message, $how, $type, @args ) = @_;
 
-   if( $how == CHANGE_PUSH ) {
-      $message->pack_all_sametype( $pdef->type, @args );
+   if( $how == CHANGE_SET ) {
+      my ( $value ) = @args;
+      $message->pack_typed( "list($type)", $value );
+   }
+   elsif( $how == CHANGE_PUSH ) {
+      $message->pack_all_sametype( $type, @args );
    }
    elsif( $how == CHANGE_SHIFT ) {
       my ( $count ) = @args;
@@ -641,7 +555,7 @@ sub _generate_message_UPDATE_array
       my ( $start, $count, @values ) = @args;
       $message->pack_int( $start );
       $message->pack_int( $count );
-      $message->pack_all_sametype( $pdef->type, @values );
+      $message->pack_all_sametype( $type, @values );
    }
    elsif( $how == CHANGE_MOVE ) {
       my ( $index, $delta ) = @args;
@@ -656,14 +570,19 @@ sub _generate_message_UPDATE_array
 sub _generate_message_UPDATE_objset
 {
    my $self = shift;
-   my ( $message, $how, $pdef, @args ) = @_;
+   my ( $message, $how, $type, @args ) = @_;
 
-   if( $how == CHANGE_ADD ) {
+   if( $how == CHANGE_SET ) {
       my ( $value ) = @args;
-      $pdef->type->pack_value( $message, $value );
+      # This will arrive in a plain LIST ref
+      $message->pack_typed( "list($type)", $value );
+   }
+   elsif( $how == CHANGE_ADD ) {
+      my ( $value ) = @_;
+      $message->pack_typed( $type, $value );
    }
    elsif( $how == CHANGE_DEL ) {
-      my ( $id ) = @args;
+      my ( $id ) = @_;
       $message->pack_int( $id );
    }
    else {
@@ -671,10 +590,11 @@ sub _generate_message_UPDATE_objset
    }
 }
 
+# Keep perl happy; keep Britain tidy
+1;
+
+__END__
+
 =head1 AUTHOR
 
 Paul Evans <leonerd@leonerd.org.uk>
-
-=cut
-
-0x55AA;
