@@ -1,14 +1,11 @@
 #  You may distribute under the terms of either the GNU General Public License
 #  or the Artistic License (the same terms as Perl itself)
 #
-#  (C) Paul Evans, 2010-2011 -- leonerd@leonerd.org.uk
+#  (C) Paul Evans, 2010 -- leonerd@leonerd.org.uk
 
 package Tangence::Server::Context;
 
 use strict;
-use warnings;
-
-our $VERSION = '0.20_001';
 
 use Carp;
 
@@ -17,11 +14,11 @@ use Tangence::Constants;
 sub new
 {
    my $class = shift;
-   my ( $stream, $token ) = @_;
+   my ( $conn, $token ) = @_;
 
    return bless {
-      stream => $stream,
-      token  => $token,
+      conn  => $conn,
+      token => $token,
    }, $class;
 }
 
@@ -31,10 +28,10 @@ sub DESTROY
    $self->{responded} or croak "$self never responded";
 }
 
-sub stream
+sub connection
 {
    my $self = shift;
-   return $self->{stream};
+   return $self->{conn};
 }
 
 sub respond
@@ -44,7 +41,8 @@ sub respond
 
    $self->{responded} and croak "$self has responded once already";
 
-   $self->stream->respond( $self->{token}, $message );
+   my $conn = $self->{conn};
+   $conn->respond( $self->{token}, $message );
 
    $self->{responded} = 1;
 
@@ -56,15 +54,16 @@ sub responderr
    my $self = shift;
    my ( $msg ) = @_;
 
-   $self->respond( Tangence::Message->new( $self->stream, MSG_ERROR )
+   $self->respond( Tangence::Message->new( $self->{conn}, MSG_ERROR )
       ->pack_str( $msg )
    );
 }
 
+# Keep perl happy; keep Britain tidy
+1;
+
+__END__
+
 =head1 AUTHOR
 
 Paul Evans <leonerd@leonerd.org.uk>
-
-=cut
-
-0x55AA;
